/** Solver - instruction generator
 *
 * @author Ondrej Cekan
 * @date 20/02/2014
 * @update 05/07/2015
 *
     Copyright (C) 2015 Ondrej Cekan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <iostream>
#include <fstream>


/** Solver class
 */
class Solver
{
    unsigned i = 0;                             // cycle variable
    unsigned LIMAX;                             // max instructions

    ofstream fout;                              // output file
    vector<pair<Tsyn *,string> > nextgensyn;    // buffer of instructions

    bool is_end = false;                        // boolean value for signalizing that required number of instructions was reached
    bool is_random = false;

    string uosyntax;                            // random VLIW part of syntax
    Tsyn *uosyn = NULL;                         // useonly syntax
    int uoslot = 0;                             // actual processed slot

    Tsyn cons_string;                           // syntax of string


public:
        /** Constructor
         *
         * @param output_file - output file name
         * @param maxlines - number of instructions
         *
         */
        Solver (string output_file, unsigned maxlines)
        {
            fout.open (output_file.c_str());
            cons_string.name = ":string:";

            LIMAX = maxlines;

            // generate start() items
            cons_startend("start", &cons_string, &nextgensyn);
        }

        /** Generate all instructions
         */
        void generate_all()
        {
            get(true);
        }

        /** Generate one instruction
         */
        void generate()
        {
            get(false);
        }

        /** Generator
         *
         * @param all - choose between one or all instructions generation
         *
         */
        void get(bool all)
        {
            while ((i <= LIMAX && LIMAX < UNSIGNED_MAX) || (nextgensyn.size() > 0) || !is_end)
            { // generating of instructions
                i++;
                bool ignore_prob = false;               // ignore probability
                bool ignore_composed = false;           // ignore composed instruction
                pair<string, unsigned> store;           // store current position

                Tsyn *syntax = NULL;               // actually processed syntax
                string flag = "";                  // flag for random instruction in constraint

                if (nextgensyn.size() > 0)
                { // selection from buffer of instructions
                    ignore_prob = true;
                    ignore_composed = true;

                    syntax = (nextgensyn.begin())->first;
                    flag = (nextgensyn.begin())->second;

                    if (syntax->name.compare(":string:") == 0)
                    { // string swap into syntax
                        vector<pair<string, string> >::iterator valIt = syntax->val.begin();
                        syntax->syntax = valIt->first;
                        syntax->val.erase(syntax->val.begin());
                    }

                    nextgensyn.erase(nextgensyn.begin());
                }
                else
                { // random selection of instruction

                    if (!is_random)
                    {
                        int useonlynum = 0;
                        for (gconsIt = gcons.begin(); gconsIt < gcons.end(); gconsIt++)
                        {
                            if (gconsIt->op.compare("useonly") == 0)
                            {
                                useonlynum++;
                            }
                        }

                        int usechoose; // number of useonly instruction for choose
                        if (useonlynum == 0) usechoose = 0;
                        else usechoose = rand() % useonlynum;

                        for (gconsIt = gcons.begin(); gconsIt < gcons.end(); gconsIt++)
                        {
                            if (gconsIt->op.compare("useonly") == 0)
                            {
                                if (usechoose == 0)
                                {
                                    ignore_prob = true;
                                    syntax = gconsIt->syn;
                                    uosyn = gconsIt->syn;
                                    uoslot = -1;
                                    break;
                                }
                                usechoose--;
                            }
                        }
                    }

                    if (syntax == NULL) syntax = get_random_gen(syn.begin(),syn.size());
                }

                // taking syntax for operations
                string synt = syntax->syntax;

                // check slot
                if (uosyn != NULL && uoslot >= 0)
                {
                    bool fineslot = true;

                    for (syntax->consIt = syntax->cons.begin(); syntax->consIt < syntax->cons.end(); syntax->consIt++)
                    {
                        if (syntax->consIt->op.compare("slot") == 0)
                        {
                            if (&(*syntax->consIt->syn) == &(*uosyn) || syntax->consIt->str.compare("all") == 0)
                            {
                                fineslot = false;
                                if (syntax->consIt->rto.compare(intToString(uoslot)) == 0)
                                {
                                    fineslot = true;
                                    break;
                                }
                            }
                        }
                    }

                    if (!fineslot) { i--; continue; }
                }

                // probability constraints
                if (!ignore_prob && !cons_probability(syntax,ignore_prob)) { i--; continue; }

                // extended probability for vliw
                if (ignore_prob && uoslot == -1 && !cons_probability(syntax,ignore_prob)) { i--; continue; }

                // beforeinsert() constraint
                if (flag.compare("noib") != 0)
                {
                    vector<pair<Tsyn *,string> > items = cons_beforeinsert(syntax);
                    if (items.size() > 1)
                    {
                        for (vector<pair<Tsyn *,string> >::reverse_iterator itemsIt = items.rbegin(); itemsIt < items.rend(); itemsIt++)
                        {
                            if ((itemsIt->first) == &(*syntax))
                            {
                                itemsIt->second = "noib"; // no insert before for this value in buffer, disables executing more than one same instruction
                            }

                            nextgensyn.insert(nextgensyn.begin(), *itemsIt);
                        }

                        i--;
                        continue;
                    }
                }

                // substitution
                if (!cons_substitute(&subs, &rate, &synt, flag, &(syntax->cons), i, &fout)) { i--;  continue; }

                for (syntax->consIt = syntax->cons.begin(); syntax->consIt < syntax->cons.end(); syntax->consIt++)
                {
                    if (syntax->consIt->op.compare("annulcomposed") == 0)
                    {
                        ignore_composed = true;
                        break;
                    }
                }

                // storeval nocomposed
                bool storevalerr = false;
                if (!ignore_composed)
                {
                    for (storevalIt = storeval.begin(); storevalIt < storeval.end(); storevalIt++)
                    {
                        if (storevalIt->first.compare("nocomposed") == 0)
                        {
                            int nol = numOfLines(synt);

                            unsigned apos = IMAX-(LIMAX-nol-i)-CPBALANCE;

                            if (nol > 0 && apos >= storevalIt->second)
                            {
                                storevalerr = true;
                                break;
                            }
                        }
                    }
                }
                if (storevalerr) { i--; continue; }

                for (storevalIt = storeval.begin(); storevalIt < storeval.end(); storevalIt++)
                {
                    if (storevalIt->first.compare("nocomposed") == 0)
                    {
                        int nol = numOfLines(synt);

                        unsigned apos = IMAX-(LIMAX-nol-i)-CPBALANCE;

                        if (apos >= storevalIt->second)
                        {
                            storeval.erase(storevalIt);
                        }
                    }
                }

                // storing cpos for stoteval - it must be done before variable selection
                for (syntax->consIt = syntax->cons.begin(); syntax->consIt < syntax->cons.end(); syntax->consIt++)
                {
                    if (syntax->consIt->op.compare("storeval") == 0)
                    {
                        if (syntax->consIt->rfrom.compare("cpos") == 0)
                        {
                            int nol = numOfLines(synt);

                            store.first = syntax->consIt->str;
                            store.second += IMAX-(LIMAX-nol-i)-CPBALANCE;
                        }

                        if (syntax->consIt->rto.compare("cpos") == 0)
                        {
                            int nol = numOfLines(synt);

                            store.first = syntax->consIt->str;
                            store.second += IMAX-(LIMAX-nol-i)-CPBALANCE;
                        }
                    }
                }


                ////////////////////////////////////////////////////////////////// variable selection and substitution
                for (varIt = var.begin(); varIt < var.end() && syntax->name.compare(":string:") != 0; varIt++)
                { // browsing all variables
                    if (synt.find(varIt->name) != string::npos)
                    { // variable exists in syntax form
                        vector<Tconsit>::iterator deleteIt;
                        vector<pair<string, string> >::iterator deleteitem;
                        bool uniq = false;
                        string num;

                        for (syntax->consIt = syntax->cons.begin(); syntax->consIt < syntax->cons.end(); syntax->consIt++)
                        {
                            if (syntax->consIt->op.compare("contain") == 0)
                            { // contain() constraint
                                if (varIt->name == syntax->consIt->rfrom)
                                {
                                    switch (syntax->consIt->type)
                                    {
                                        case CSYN:
                                        {
                                            if (syntax->consIt->syn->val.size() > 0)
                                            {
                                                vector<vector<pair<string, string> >::iterator> val; // vector of all possible substitutes

                                                for (vector<pair<string, string> >::iterator valIt = syntax->consIt->syn->val.begin(); valIt < syntax->consIt->syn->val.end(); valIt++)
                                                { // browsing all possible substitutes from another instruction
                                                    if (valIt->first == syntax->consIt->rto)
                                                    {
                                                        val.insert(val.begin(), valIt);
                                                    }
                                                }

                                                if (val.size() > 0)
                                                {
                                                    vector<pair<string, string> >::iterator valIt = val.at(rand() % val.size());
                                                    if (num.empty() || rand()%2 == 1)
                                                    { // random choosing one value or new one
                                                        num.assign(valIt->second);
                                                        deleteIt = (syntax->consIt);
                                                        deleteitem = valIt;
                                                    }
                                                }
                                            }
                                        } break;
                                        default: break;
                                    }
                                }
                            }
                            else if (syntax->consIt->op.compare("unique") == 0)
                            { // unique() constraint
                                uniq = true;
                            }
                        }


                        unsigned vmin = 0;  // default min value
                        unsigned vmax = 0;  // default max value

                        // max values based on type
                        if (varIt->type == VINT)        vmax = 256;
                        else if (varIt->type == VINT14) vmax = 16384;
                        else if (varIt->type == VINT16) vmax = 65536;

                        // constraint for variable
                        for (varIt->consIt = varIt->cons.begin(); varIt->consIt < varIt->cons.end(); varIt->consIt++)
                        {
                            if (varIt->consIt->op.compare("min") == 0)
                            {
                                if (varIt->consIt->str.compare("cpos") == 0)
                                {
                                    int nol = numOfLines(synt);

                                    vmin = IMAX-(LIMAX-nol-i)-CPBALANCE;
                                }
                                else
                                {
                                    vmin = varIt->consIt->val;
                                }
                            }
                            else if (varIt->consIt->op.compare("max") == 0)
                            {
                                int nol = numOfLines(synt);

                                if (varIt->consIt->str.compare("cpos") == 0)
                                {
                                    vmax = IMAX-(LIMAX-nol-i)-CPBALANCE;
                                }
                                else if (varIt->consIt->str.compare("rest") == 0)
                                {
                                    vmax = varIt->consIt->val;

                                    //MAXLINES - CURRENT POSITION
                                    if (vmax > (unsigned)abs((int)(LIMAX-nol-i)))
                                        vmax = (unsigned)abs((int)(LIMAX-nol-i));
                                }
                                else
                                {
                                    vmax = varIt->consIt->val;
                                }
                            }
                        }

                        if (vmax < vmin) vmax = vmin+vmax;


                        string str;
                        if (varIt->type == VINT ||
                            varIt->type == VINT14 ||
                            varIt->type == VINT16)
                        { // number data type
                            int inum;
                            bool cycle;

                            do
                            {
                                inum = (rand()%(vmax+1-vmin))+vmin;
                                cycle = false;

                                for (varIt->consIt = varIt->cons.begin(); varIt->consIt < varIt->cons.end(); varIt->consIt++)
                                {
                                    if (varIt->consIt->op.compare("mdiv") == 0)
                                    { // mdiv() constraint
                                        if (inum%(varIt->consIt->val) != 0) cycle = true;
                                    }
                                }
                            } while(cycle);

                            for (vector<Tconsit>::reverse_iterator consItr = varIt->cons.rbegin(); consItr < varIt->cons.rend(); consItr++)
                            {
                                if (consItr->op.compare("mul") == 0)
                                { // mul() constraint
                                    inum *= consItr->val;
                                }
                                else if (consItr->op.compare("div") == 0)
                                { // div() constraint
                                    inum /= consItr->val;
                                }
                                else if (consItr->op.compare("mod") == 0)
                                { // div() constraint
                                    inum %= consItr->val;
                                }
                            }

                            for (syntax->consIt = syntax->cons.begin(); syntax->consIt < syntax->cons.end(); syntax->consIt++)
                            {
                                if (syntax->consIt->op.compare("storeval") == 0)
                                {
                                    if (&(*syntax->consIt->var) == &(*varIt))
                                    {
                                        store.first = syntax->consIt->str;
                                        store.second += (inum/syntax->consIt->val);
                                    }

                                    if (&(*syntax->consIt->var2) == &(*varIt))
                                    {
                                        store.first = syntax->consIt->str;
                                        store.second += (inum/syntax->consIt->val2);
                                    }
                                }
                            }

                            num = intToString(inum);
                        }
                        else if (varIt->type == VSTR)
                        { // string data type = STR + number (e.g. STR1234)
                            str = "STR";

                            if (num.empty())
                            { // random number for string
                                num = intToString(rand()%30000);
                            }
                            else
                            { // number is assigned sometimes before (e.g. via contain() constraint)
                                deleteIt->syn->eval.insert(deleteIt->syn->eval.begin(), make_pair(deleteitem->first,deleteitem->second));
                                deleteIt->syn->val.erase(deleteitem); // delete
                            }

                            int i = 0;
                            bool found = false;
                            do
                            {
                                i++;
                                if (i > 1) num = intToString(rand()%30000); // string with this number cannot be assigned because of any constraint, choose new one

                                found = false;


                                // check if string is not in buffer with  upcoming values
                                for (vector<pair<string, string> >::iterator valIt = syntax->val.begin(); valIt < syntax->val.end() && !found; valIt++)
                                {
                                    if (valIt->first.compare(varIt->name) == 0 && valIt->second.compare(num) == 0)
                                    {
                                        found = true;
                                    }
                                }

                                // check if string is not in buffer with generated values
                                for (vector<pair<string, string> >::iterator valIt = syntax->eval.begin(); valIt < syntax->eval.end() && !found; valIt++)
                                {
                                    if (valIt->first.compare(varIt->name) == 0 && valIt->second.compare(num) == 0)
                                    {
                                        found = true;
                                    }
                                }


                                // check other possible conditions for this assign value
                                for (syntax->consIt = syntax->cons.begin(); syntax->consIt < syntax->cons.end() && !found; syntax->consIt++)
                                {
                                    if (syntax->consIt->op.compare("different") == 0)
                                    {
                                        switch (syntax->consIt->type)
                                        {
                                            case CSYN:
                                            {
                                                for (vector<pair<string, string> >::iterator valIt = syntax->consIt->syn->val.begin(); valIt < syntax->consIt->syn->val.end() && !found; valIt++)
                                                {
                                                    if (valIt->first.compare(varIt->name) == 0 && valIt->second.compare(num) == 0)
                                                    {
                                                        found = true;
                                                    }
                                                }

                                                for (vector<pair<string, string> >::iterator valIt = syntax->consIt->syn->eval.begin(); valIt < syntax->consIt->syn->eval.end() && !found; valIt++)
                                                {
                                                    if (valIt->first.compare(varIt->name) == 0 && valIt->second.compare(num) == 0)
                                                    {
                                                        found = true;
                                                    }
                                                }
                                            } break;
                                            default: break;
                                        }
                                    }
                                }

                                // if all right insert into buffer with generated values
                                if (!found) syntax->val.insert(syntax->val.begin(), make_pair(varIt->name,num));

                            } while (found && uniq);
                        }

                        // replace
                        synt = replace_str(synt, varIt->name, str+num);
                    }
                }

                for (syntax->consIt = syntax->cons.begin(); syntax->consIt < syntax->cons.end(); syntax->consIt++)
                {
                    if (syntax->consIt->op.compare("nolinecount") == 0)
                    {
                        i--;
                        break;
                    }
                }

                // cpos replacing
                synt = replace_str(synt, "cpos", intToString(IMAX-(LIMAX-numOfLines(synt)-i)-CPBALANCE));

                if (!store.first.empty())
                { // storing CPOS to storeval
                    storeval.insert(storeval.begin(),store);
                }

                if (synt.find("random")!=string::npos || is_random) // 1 cycle more
                { // random keyword replacing, copying and composed useonly instruction building
                    uosyntax = replace_one(uosyntax, "random", synt);
                    if (!is_random) uosyntax.assign(synt);

                    if (uosyntax.find("random")!=string::npos)
                    {
                        vector<string> x = split(uosyntax, '\n');

                        int uox = 0;
                        for (vector<string>::iterator xIt = x.begin(); xIt < x.end(); xIt++)
                        {
                            if (xIt->find("random")!=string::npos)
                            {
                                uoslot = uox%NSLOTS;
                                break;
                            }
                            uox++;
                        }

                        is_random = true;
                        i--;
                        continue;
                    }
                    else
                    {
                        LIMAX++;
                        uosyn = NULL;
                        is_random = false;
                    }
                }

                if (!is_random)
                { // all random keywords are replaced, finish this instruction
                    if (!uosyntax.empty())
                    { // copying
                        synt.assign(uosyntax);
                        uosyntax.erase();
                    }
                }

                LIMAX -= numOfLines(synt);

                /// PRINT INSTRUCTION
                fout << synt << endl;
                cout << synt << endl;


                // end of generation
                if (i >= LIMAX && !is_end)
                {
                    is_end = true;

                    // generate items in buffer with upcoming values that must be written
                    for (gconsItr = gcons.rbegin(); gconsItr < gcons.rend(); gconsItr++)
                    {
                        if (gconsItr->op.compare("outall") == 0)
                        { // outall() constraint
                            switch (gconsItr->type)
                            {
                                case CSYN:
                                {
                                    for (vector<Tconsit>::iterator consIt = gconsItr->syn->cons.begin(); consIt < gconsItr->syn->cons.end(); consIt++)
                                    {
                                        if (consIt->op.compare("contain") == 0)
                                        {
                                            for (vector<pair<string, string> >::iterator valIt = consIt->syn->val.begin(); valIt < consIt->syn->val.end(); valIt++)
                                            {
                                                nextgensyn.push_back(make_pair(gconsItr->syn,""));
                                            }
                                        }
                                    }
                                } break;
                                default: break;
                            }
                        }
                    }

                    // generate end() items
                    cons_startend("end", &cons_string, &nextgensyn);
                }

                // afterinsert() constraint
                for (syntax->consIt = syntax->cons.begin(); syntax->consIt < syntax->cons.end(); syntax->consIt++)
                {
                    if (syntax->consIt->op.compare("afterinsert") == 0)
                    {
                        switch (syntax->consIt->type)
                        {
                            case CSTR:
                            { // string
                                if (!syntax->consIt->str.empty())
                                {
                                    cons_string.val.insert(cons_string.val.begin(), make_pair(syntax->consIt->str,""));
                                    nextgensyn.insert(nextgensyn.begin(), make_pair(&cons_string,""));
                                }
                            } break;
                            case CSYN:
                            { // instruction
                                nextgensyn.insert(nextgensyn.begin(), make_pair(syntax->consIt->syn,""));
                            } break;
                            default: break;
                        }
                    }
                }

                if (!all) break;
            }
        }
};

